#!/usr/bin/env python
# Copyright 2023 NAREF Development Team
#
# Unless required by applicable law or agreed to in writing, this software
# is on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied.

"""A quantum & classical reservoir handler for Temporal Series Prediction."""

from naref.modules.comp import Comparator
from naref.modules.quantum.hyper import HyperQ
from naref.modules.classical.hyper import HyperC
from naref.modules.demo import fastest_hs
from naref.modules import *
